import pytest
import asyncio
from unittest.mock import AsyncMock, patch
import argparse
import openai

import coverup.llm as llm


@pytest.mark.asyncio
async def test_do_chat_authentication_error():
    class MockCodeSegment:
        def __init__(self, name):
            self.name = name

    class MockResponse:
        def __init__(self):
            self.request = None
            self.status_code = 401
            self.headers = {}

    seg = MockCodeSegment("test_segment")
    messages = [{'role': 'assistant',
                 'content': 'test prompt'}]

    with patch("litellm.acreate",
               new=AsyncMock(side_effect=openai.AuthenticationError("Authentication failed", response=MockResponse(), body=''))):
        with patch.object(llm.Chatter, '_validate_model'):
            chatter = llm.Chatter(model="gpt-4")

            with pytest.raises(openai.AuthenticationError):
                await chatter.chat(messages)



import typing as T
from .prompter import Prompter, CodeSegment, mk_message, get_module_name
from ..utils import lines_branches_do


class ClaudePrompter(Prompter):
    """Prompter for Claude."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def initial_prompt(self, segment: CodeSegment) -> T.List[dict]:
        module_name = get_module_name(segment.path, self.args.src_base_dir)
        filename = segment.path.relative_to(self.args.src_base_dir)

        return [
            mk_message("You are an expert Python test-driven developer who creates pytest test functions that achieve high coverage.",
                    role="system"),
            mk_message(f"""
<file path="{filename}" module_name="{module_name}">
{segment.get_excerpt(tag_lines=bool(segment.executed_lines))}
</file>

<instructions>

The code above does not achieve full coverage:
when tested, {'it does' if not segment.executed_lines else segment.lines_branches_missing_do()} not execute.

1. Create a new pytest test function that executes these missing lines/branches, always making
sure that the new test is correct and indeed improves coverage.

2. Always send entire Python test scripts when proposing a new test or correcting one you
previously proposed.

3. Be sure to include assertions in the test that verify any applicable postconditions.

4. Please also make VERY SURE to clean up after the test, so as not to affect other tests;
use 'pytest-mock' if appropriate.

5. Write as little top-level code as possible, and in particular do not include any top-level code
calling into pytest.main or the test itself.

6.  Respond with the Python code enclosed in backticks. Before answering the question, please think about it step-by-step within <thinking></thinking> tags. Then, provide your final answer within <answer></answer> tags.
</instructions>
""")
        ]


    def error_prompt(self, segment: CodeSegment, error: str) -> T.List[dict]:
        return [mk_message(f"""\
<error>{error}</error>
Executing the test yields an error, shown above.
<instructions>
1. Modify the test to correct it.
2. Respond with the complete Python code in backticks.
3. Before answering the question, please think about it step-by-step within <thinking></thinking> tags. Then, provide your final answer within <answer></answer> tags.
</instructions>
""")
        ]


    def missing_coverage_prompt(self, segment: CodeSegment,
                                missing_lines: set, missing_branches: set) -> T.List[dict]:
        return [mk_message(f"""\
This test still lacks coverage: {lines_branches_do(missing_lines, set(), missing_branches)} not execute.
<instructions>
1. Modify it to execute those lines.
2. Respond with the complete Python code in backticks.
3. Before responding, please think about it step-by-step within <thinking></thinking> tags. Then, provide your final answer within <answer></answer> tags.
</instructions>
""")
        ]



from pathlib import Path
import tempfile
import subprocess
import pytest
import typing as T
import sys
import json
import os
from .utils import subprocess_run


async def measure_test_coverage(*, test: str, tests_dir: Path, pytest_args='',
                                log_write=None, isolate_tests=False, branch_coverage=True):
    """Runs a given test and returns the coverage obtained."""
    with tempfile.NamedTemporaryFile(prefix="tmp_test_", suffix='.py', dir=str(tests_dir), mode="w") as t:
        t.write(test)
        t.flush()

        with tempfile.NamedTemporaryFile(suffix='.json', delete=False) as j:
            try:
                # -qq to cut down on tokens
                p = await subprocess_run([sys.executable, '-m', 'slipcover',  *(('--branch',) if branch_coverage else ()),
                                          '--json', '--out', j.name,
                                          '-m', 'pytest', *pytest_args.split(),
                                          *(('--cleanslate',) if isolate_tests else ()),
                                          '-qq', '-x', '--disable-warnings', t.name],
                                         check=True, timeout=60)
                if log_write:
                    log_write(str(p.stdout, 'UTF-8', errors='ignore'))

                # not checking for JSON errors here because if pytest aborts, its RC ought to be !=0
                cov = json.load(j)
            finally:
                j.close()
                try:
                    os.unlink(j.name)
                except FileNotFoundError:
                    pass

    return cov


def measure_suite_coverage(*, tests_dir: Path, source_dir: T.Optional[Path], pytest_args='',
                           trace=None, isolate_tests=False, branch_coverage=True):
    """Runs an entire test suite and returns the coverage obtained."""

    with tempfile.NamedTemporaryFile(suffix='.json', delete=False) as j:
        try:
            command = [sys.executable,
                       '-m', 'slipcover',
                             *(('--source', source_dir) if source_dir else ()),
                             *(('--branch',) if branch_coverage else ()),
                             '--json', '--out', j.name,
                       '-m', 'pytest', *pytest_args.split(), *(('--cleanslate',) if isolate_tests else ()),
                             '--disable-warnings', '-x', tests_dir]

            if trace: trace(command)
            p = subprocess.run(command, check=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            if p.returncode not in (pytest.ExitCode.OK, pytest.ExitCode.NO_TESTS_COLLECTED):
                if trace: trace(f"tests rc={p.returncode}\n" + str(p.stdout, 'utf-8'))
                p.check_returncode()

            try:
                return json.load(j)
            except json.decoder.JSONDecodeError:
                # The JSON is broken, so pytest's execution likely aborted (e.g. a Python unhandled exception).
                p.check_returncode() # this will almost certainly raise an exception. If not, we do it ourselves:
                raise subprocess.CalledProcessError(p.returncode, command, output=p.stdout)
        finally:
            j.close()

            try:
                os.unlink(j.name)
            except FileNotFoundError:
                pass



# file src/coverup/coverup.py:323-325
# lines [325]
# branches []

import pytest
from unittest.mock import Mock
from coverup.coverup import State

@pytest.fixture
def state():
    initial_coverage = {"line1": True, "line2": False}
    state = State(initial_coverage)
    return state

def test_get_initial_coverage(state):
    expected_coverage = {"line1": True, "line2": False}
    assert state.get_initial_coverage() == expected_coverage



from pathlib import Path
import coverup.segment as segment
import textwrap
import json


somecode_py = textwrap.dedent("""\
    # Sample Python code used to create some tests.
    import os

    class Foo:
        '''docstring...'''

        @staticmethod
        def foo():
            return os.path.exists('here')

        def __init__(self):
            '''initializes...'''
            self._foo = 0

        x = 0
        if x != 0:
            y = 2

        class Bar:
            z = 10

        def bar():
            '''docstring'''
            class Baz:
                assert False

            def baz():
                pass

    GLOBAL = 42

    def func(x):
        '''docstring'''
        if x > 0:
            return 42**42

    if __name__ == "__main__":
        ...
""")

somecode_json = """\
{
    "files": {
        "tests/somecode.py": {
            "executed_lines": [
                2, 4, 5, 7, 8, 11, 15, 16, 19,
                20, 22, 30, 32, 37, 38
            ],
            "missing_lines": [
                9, 13, 17, 24, 25, 27, 28, 34, 35
            ],
            "executed_branches": [
                [ 16, 19 ], [ 37, 38 ]
            ],
            "missing_branches": [
                [ 16, 17 ],
                [ 34, 0 ],
                [ 34, 35 ],
                [ 37, 0 ]
            ]
        }
    }
}
"""


def test_large_limit_whole_class(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    (tmp_path / "tests").mkdir()
    (tmp_path / "tests" / "somecode.py").write_text(somecode_py)

    coverage = json.loads(somecode_json)
    segs = segment.get_missing_coverage(coverage, line_limit=100)

    assert ['Foo', 'func'] == [seg.name for seg in segs]
    assert all([Path(seg.filename).name == 'somecode.py' for seg in segs])
    assert all(seg.begin < seg.end for seg in segs)

    assert textwrap.dedent(segs[0].get_excerpt(tag_lines=False)) == textwrap.dedent("""\
        import os
        class Foo:
            '''docstring...'''

            @staticmethod
            def foo():
                return os.path.exists('here')

            def __init__(self):
                '''initializes...'''
                self._foo = 0

            x = 0
            if x != 0:
                y = 2

            class Bar:
                z = 10

            def bar():
                '''docstring'''
                class Baz:
                    assert False

                def baz():
                    pass
        """)

    assert textwrap.dedent(segs[1].get_excerpt(tag_lines=False)) == textwrap.dedent("""\
        def func(x):
            '''docstring'''
            if x > 0:
                return 42**42
        """)

    # FIXME check executed_lines, missing_lines, ..., interesting_lines


def test_small_limit(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    (tmp_path / "tests").mkdir()
    (tmp_path / "tests" / "somecode.py").write_text(somecode_py)

    coverage = json.loads(somecode_json)
    segs = segment.get_missing_coverage(coverage, line_limit=3)

    # "Bar" omitted because executed
    assert ['foo', '__init__', 'bar', 'func'] == [seg.name for seg in segs]
    assert all([Path(seg.filename).name == 'somecode.py' for seg in segs])
    assert all(seg.begin < seg.end for seg in segs)


    assert textwrap.dedent(segs[0].get_excerpt(tag_lines=False)) == textwrap.dedent("""\
        import os
        class Foo:
            @staticmethod
            def foo():
                return os.path.exists('here')
        """)

    assert textwrap.dedent(segs[2].get_excerpt(tag_lines=False)) == textwrap.dedent("""\
        class Foo:
            def bar():
                '''docstring'''
                class Baz:
                    assert False

                def baz():
                    pass
        """)

    # FIXME check executed_lines, missing_lines, ..., interesting_lines


def test_all_missing_not_loaded(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    (tmp_path / "tests").mkdir()
    (tmp_path / "tests" / "somecode.py").write_text(somecode_py)

    neverloaded_json = """\
        {
            "files": {
                "tests/somecode.py": {
                    "executed_lines": [],
                    "missing_lines": [
                        2, 4, 5, 7, 8, 9, 11, 13, 15, 16, 17,
                        19, 20, 22, 24, 25, 27, 28, 30, 32, 34,
                        35, 37, 38 ],
                    "executed_branches": [],
                    "missing_branches": [
                        [ 16, 17 ], [ 16, 19 ], [ 34, 0 ], [ 34, 35 ],
                        [ 37, 0 ], [ 37, 38 ]
                    ]
                }
            }
        }"""

    coverage = json.loads(neverloaded_json)
    segs = segment.get_missing_coverage(coverage, line_limit=2)

    assert ['foo', '__init__', 'Bar', 'bar', 'func'] == [seg.name for seg in segs]
    assert all([Path(seg.filename).name == 'somecode.py' for seg in segs])
    assert all(seg.begin < seg.end for seg in segs)

    assert textwrap.dedent(segs[3].get_excerpt(tag_lines=False)) == textwrap.dedent("""\
        class Foo:
            def bar():
                '''docstring'''
                class Baz:
                    assert False

                def baz():
                    pass
        """)

    # FIXME check executed_lines, missing_lines, ..., interesting_lines


def test_only_branch_missing(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    (tmp_path / "code.py").write_text(textwrap.dedent("""\
        class Foo:
            class Bar:
                def __init__(self, x):
                    self.x = None
                    if x:
                        self.x = x
                    self.y = 0

        """))

    code_json = """\
        {
            "files": {
                "code.py": {
                    "executed_lines": [
                        1, 2, 3, 4, 5, 6, 7
                    ],
                    "missing_lines": [
                    ],
                    "executed_branches": [
                        [5, 6]
                    ],
                    "missing_branches": [
                        [5, 7]
                    ]
                }
            }
        }"""

    coverage = json.loads(code_json)
    segs = segment.get_missing_coverage(coverage, line_limit=4)

    assert len(segs) == 1
    assert "__init__" == segs[0].name

    assert textwrap.dedent(segs[0].get_excerpt()) == textwrap.dedent("""\
           class Foo:
               class Bar:
                   def __init__(self, x):
                       self.x = None
        5:             if x:
                           self.x = x
        7:             self.y = 0
        """)




from pathlib import Path
import typing as T
import subprocess


def format_ranges(lines: T.Set[int], negative: T.Set[int]) -> str:
    """Formats sets of line numbers as comma-separated lists, collapsing neighboring lines into ranges
       for brevity."""

    def get_range(lines):
        it = iter(sorted(lines))

        a = next(it, None)
        while a is not None:
            b = a
            while (n := next(it, None)) is not None and not (set(range(b+1,n+1)) & negative):
                b = n

            if a == b:
                yield str(a)
            else:
                yield f"{a}-{b}"

            a = n

    return ", ".join(get_range(lines))


def format_branches(branches):
    for br in sorted(branches):
        yield f"{br[0]}->exit" if br[1] == 0 else f"{br[0]}->{br[1]}"


def lines_branches_do(lines: T.Set[int], neg_lines: T.Set[int], branches: T.Set[T.Tuple[int, int]]) -> str:
    relevant_branches = {b for b in branches if b[0] not in lines and b[1] not in lines} if branches else set()

    s = ''
    if lines:
        s += f"line{'s' if len(lines)>1 else ''} {format_ranges(lines, neg_lines)}"

        if relevant_branches:
            s += " and "

    if relevant_branches:
        s += f"branch{'es' if len(relevant_branches)>1 else ''} "
        s += ", ".join(format_branches(relevant_branches))

    s += " does" if len(lines)+len(relevant_branches) == 1 else " do"
    return s


async def subprocess_run(args: T.Sequence[str], check: bool = False, timeout: T.Optional[int] = None) -> subprocess.CompletedProcess:
    """Provides an asynchronous version of subprocess.run"""
    import asyncio

    process = await asyncio.create_subprocess_exec(*args, stdout=asyncio.subprocess.PIPE,
                                                   stderr=asyncio.subprocess.STDOUT)

    try:
        if timeout is not None:
            output, _ = await asyncio.wait_for(process.communicate(), timeout=timeout)
        else:
            output, _ = await process.communicate()

    except (asyncio.TimeoutError, asyncio.exceptions.TimeoutError):
        try:
            process.terminate()
            await process.wait()
        except ProcessLookupError:
            pass

        if timeout:
            timeout_f = float(timeout)
        else:
            timeout_f = 0.0
        raise subprocess.TimeoutExpired(args, timeout_f) from None

    if check and process.returncode:
        raise subprocess.CalledProcessError(process.returncode, args, output=output)

    # process.returncode is None iff the process hasn't terminated yet
    return subprocess.CompletedProcess(args=args, returncode=T.cast(int, process.returncode), stdout=output)


def summary_coverage(cov: dict, sources: T.List[Path]) -> str:
    """Returns the summary coverage, limiting it to the source files if any are given."""

    if sources:
        import copy
        from slipcover.slipcover import add_summaries

        filtered = {
            'meta': cov['meta'],
            'files': {}
        }

        for f in cov['files']:
            resolved = Path(f).resolve()
            if resolved in sources:
                filtered['files'][str(resolved)] = copy.deepcopy(cov['files'][f])

        add_summaries(filtered)
        cov = filtered

    return f"{cov['summary']['percent_covered']:.1f}%"



# file src/coverup/coverup.py:254-256
# lines [256]
# branches []

import pytest
import typing as T
from unittest.mock import patch, MagicMock
from coverup.coverup import get_required_modules

@pytest.fixture
def mock_module_available():
    return {
        'module1': 1,
        'module2': 0,
        'module3': 1,
        'module4': 0
    }

def test_get_required_modules(mock_module_available):
    with patch('coverup.coverup.module_available', new=mock_module_available):
        result = get_required_modules()
        expected = ['module2', 'module4']
        assert sorted(result) == sorted(expected)



import pytest
from pathlib import Path
from coverup.segment import CodeSegment
from coverup.prompt.prompter import get_module_name
from coverup.prompt.gpt_v1 import GptV1Prompter
from coverup.prompt.gpt_v2_ablated import GptV2Prompter, GptV2AblatedPrompter
from coverup.prompt.gpt_v2_fully_ablated import GptV2FullyAblatedPrompter
from coverup.prompt.claude import ClaudePrompter
from coverup.coverup import parse_args
import coverup.codeinfo


def test_get_module_name():
    fpath = Path('src/flask/json/provider.py').resolve()
    base_path = Path('src').resolve()

    assert 'flask.json.provider' == get_module_name(fpath, base_path)

    assert None == get_module_name(fpath, Path('./tests').resolve())


class MockSegment(CodeSegment):
    def __init__(self):
        super().__init__(
            filename=Path("lib") / "ansible" / "context.py",
            name="foo",
            begin=10, end=20, 
            lines_of_interest=set(),
            missing_lines=set(),
            executed_lines=set(),
            missing_branches=set(),
            context=[],
            imports=[]
        )

    def lines_branches_missing_do(self) -> str:
        return "mock lines and branches do"


    def get_excerpt(self, *, tag_lines=True, add_imports=True) -> str:
        return f"""\
            fake excerpt
            {add_imports=}
            {tag_lines=}
        """


@pytest.fixture
def pkg_fixture(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)

    (tmp_path / "lib" / "ansible").mkdir(parents=True)
    (tmp_path / "lib" / "ansible" / "__init__.py").touch()
    (tmp_path / "tests").mkdir(parents=True)

    def mock_parse_file(*args, **kwargs):
        return None

    def mock_get_info(module, name, **kwargs):
        return f"get_info {name=} generate_imports={kwargs.get('generate_imports', True)}"

    monkeypatch.setattr(coverup.codeinfo, 'parse_file', mock_parse_file)
    monkeypatch.setattr(coverup.codeinfo, 'get_info', mock_get_info)


def test_gpt_v1_relative_file_name(pkg_fixture):
    args = parse_args(["--source", "lib/ansible", "--tests", "tests", "--model", "gpt-4o"])
    print(f"{args.src_base_dir=}")

    p = GptV1Prompter(args)
    initial = p.initial_prompt(MockSegment())
    assert "extracted from ansible/context.py," in initial[0]['content']

    assert "when tested, it does not execute." in initial[0]['content']


def test_claude_relative_file_name(pkg_fixture):
    args = parse_args(["--source", "lib/ansible", "--tests", "tests", "--model", "anthropic/claude-3-sonnet-20240229"])

    p = ClaudePrompter(args)
    initial = p.initial_prompt(MockSegment())
    assert '<file path="ansible/context.py"' in initial[1]['content']

    assert "when tested, it does not execute." in initial[1]['content']


def test_ablated_nothing_ablated(pkg_fixture):
    args = parse_args(["--package", "lib/ansible", "--tests", "tests", "--model", "gpt-4o"])

    s = MockSegment()
    v2 = GptV2Prompter(args)
    ablated = GptV2AblatedPrompter(args)

    assert [ablated.get_info] == ablated.get_functions()
    assert v2.initial_prompt(s) == ablated.initial_prompt(s)
    assert v2.error_prompt(s, "the error") == ablated.error_prompt(s, "the error")
    assert v2.missing_coverage_prompt(s, {1,2,3}, set()) == ablated.missing_coverage_prompt(s, {1,2,3}, set())

    assert "get_info name='foo' generate_imports=True" in ablated.get_info(s, "foo")

def test_ablated_everything_ablated(pkg_fixture):
    args = parse_args(["--package", "lib/ansible", "--tests", "tests", "--model", "gpt-4o"])

    s = MockSegment()
    v2 = GptV2FullyAblatedPrompter(args)
    ablated = GptV2AblatedPrompter(
        args,
        with_coverage=False,
        with_get_info=False,
        with_imports=False,
        with_error_fixing=False
    )

    assert [] == ablated.get_functions()
    assert v2.initial_prompt(s) == ablated.initial_prompt(s)
    assert v2.error_prompt(s, "the error") == ablated.error_prompt(s, "the error")
    assert v2.missing_coverage_prompt(s, {1,2,3}, set()) == ablated.missing_coverage_prompt(s, {1,2,3}, set())


def test_ablated_coverage(pkg_fixture):
    args = parse_args(["--package", "lib/ansible", "--tests", "tests", "--model", "gpt-4o"])

    s = MockSegment()
    v2 = GptV2Prompter(args)
    ablated = GptV2AblatedPrompter(args, with_coverage=False)

    assert [ablated.get_info] == ablated.get_functions()
    assert "mock lines and branches" not in ablated.initial_prompt(s)[0]['content']
    assert "tag_lines=False" in ablated.initial_prompt(s)[0]['content']
    assert v2.error_prompt(s, "the error") == ablated.error_prompt(s, "the error")
    assert ablated.missing_coverage_prompt(s, {1,2,3}, set()) is None

    assert "get_info name='foo' generate_imports=True" in ablated.get_info(s, "foo")


def test_ablated_imports(pkg_fixture):
    args = parse_args(["--package", "lib/ansible", "--tests", "tests", "--model", "gpt-4o"])

    s = MockSegment()
    v2 = GptV2Prompter(args)
    ablated = GptV2AblatedPrompter(args, with_imports=False)

    assert [ablated.get_info] == ablated.get_functions()
    assert "mock lines and branches" in ablated.initial_prompt(s)[0]['content']
    assert "tag_lines=True" in ablated.initial_prompt(s)[0]['content']
    assert "add_imports=False" in ablated.initial_prompt(s)[0]['content']
    assert v2.error_prompt(s, "the error") == ablated.error_prompt(s, "the error")
    assert v2.missing_coverage_prompt(s, {1,2,3}, set()) == ablated.missing_coverage_prompt(s, {1,2,3}, set())

    assert "get_info name='foo' generate_imports=False" in ablated.get_info(s, "foo")


def test_ablated_get_info(pkg_fixture):
    args = parse_args(["--package", "lib/ansible", "--tests", "tests", "--model", "gpt-4o"])

    s = MockSegment()
    ablated = GptV2AblatedPrompter(args, with_get_info=False)

    assert ablated.get_functions() == []
    assert "get_info" not in ablated.initial_prompt(s)[0]['content']
    assert "mock lines and branches" in ablated.initial_prompt(s)[0]['content']
    assert "tag_lines=True" in ablated.initial_prompt(s)[0]['content']
    assert "add_imports=True" in ablated.initial_prompt(s)[0]['content']
    assert "get_info" not in ablated.error_prompt(s, "the error")[0]['content']
    assert "get_info" not in ablated.missing_coverage_prompt(s, {1,2,3}, set())[0]['content']


def test_ablated_error_fixing(pkg_fixture):
    args = parse_args(["--package", "lib/ansible", "--tests", "tests", "--model", "gpt-4o"])

    s = MockSegment()
    v2 = GptV2Prompter(args)
    ablated = GptV2AblatedPrompter(args, with_error_fixing=False)

    assert [ablated.get_info] == ablated.get_functions()
    assert v2.initial_prompt(s) == ablated.initial_prompt(s)
    assert ablated.error_prompt(s, "the error") is None
    assert v2.missing_coverage_prompt(s, {1,2,3}, set()) == ablated.missing_coverage_prompt(s, {1,2,3}, set())

    assert "get_info name='foo' generate_imports=True" in ablated.get_info(s, "foo")



import typing as T
from .gpt_v2 import *


class GptV2FullyAblatedPrompter(GptV2Prompter):
    """Fully ablated GPT prompter."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def initial_prompt(self, segment: CodeSegment) -> T.List[dict]:
        filename = segment.path.relative_to(self.args.src_base_dir)

        return [
            mk_message(f"""
You are an expert Python test-driven developer.
The code below, extracted from {filename}, does not achieve full coverage.
Create new pytest test functions that execute all lines and branches, always making
sure that each test is correct and indeed improves coverage.
Always send entire Python test scripts when proposing a new test or correcting one you
previously proposed.
Be sure to include assertions in the test that verify any applicable postconditions.
Please also make VERY SURE to clean up after the test, so as to avoid state pollution;
use 'monkeypatch' or 'pytest-mock' if appropriate.
Write as little top-level code as possible, and in particular do not include any top-level code
calling into pytest.main or the test itself.
Respond ONLY with the Python code enclosed in backticks, without any explanation.
```python
{segment.get_excerpt(tag_lines=False, add_imports=False)}
```
""")
        ]


    def error_prompt(self, segment: CodeSegment, error: str) -> T.List[dict] | None:
        return None


    def missing_coverage_prompt(self, segment: CodeSegment,
                                missing_lines: set, missing_branches: set) -> T.List[dict] | None:
        return None


    def get_functions(self) -> T.List[T.Callable]:
        return []



# file src/coverup/coverup.py:453-457
# lines []
# branches ['456->456']

import pytest
from coverup.coverup import extract_python

def test_extract_python_without_code_block():
    with pytest.raises(RuntimeError) as exc_info:
        extract_python("This is a response without a Python code block.")
    assert "Unable to extract Python code from response" in str(exc_info.value)



# file src/coverup/utils.py:49-51
# lines [50, 51]
# branches ['50->exit', '50->51']

import pytest
from coverup.utils import format_branches

@pytest.fixture
def mock_branches():
    return [(1, 2), (3, 0), (5, 6)]

def test_format_branches(mock_branches):
    formatted_branches = list(format_branches(mock_branches))
    assert formatted_branches == ['1->2', '3->exit', '5->6']



import coverup.coverup as coverup
from test_prompt import MockSegment, pkg_fixture
import subprocess
import sys
from pathlib import Path
import pytest


DUMMY_MODULE='mymodule'


@pytest.fixture
def uninstall():
    yield
    subprocess.check_call([sys.executable, "-m", "pip", "uninstall", "-y", DUMMY_MODULE])


def test_install(pkg_fixture, monkeypatch, uninstall):
    log = ""
    def mock_log_write(args, seg, m: str) -> None:
        nonlocal log
        log += m

    monkeypatch.setattr("coverup.coverup.log_write", mock_log_write)

    args = coverup.parse_args(["--package", "lib/ansible", "--tests", "tests", "--model", "gpt4o",
                               "--write-requirements-to", "reqs.txt"])

    missing = coverup.missing_imports(['mymodule'])
    assert missing

    ok = coverup.install_missing_imports(args, MockSegment(), missing)
    assert ok

    import importlib.metadata as md
    version = md.version(DUMMY_MODULE)

    assert DUMMY_MODULE in log
    assert version in log

    assert Path("reqs.txt").read_text() == f"{DUMMY_MODULE}=={version}\n"



# file src/coverup/coverup.py:166-182
# lines []
# branches ['170->177', '177->182']

import pytest
import re

# Assuming the original function is in coverup/coverup.py
from coverup.coverup import clean_error

# Test to cover branch 170->177
def test_clean_error_with_failures_section():
    error_message = "===== FAILURES ====\n" \
                    "_________________________ test_failure _________________________\n" \
                    "\n" \
                    "self = <test_module.TestClass object at 0x7f4b2d2c1f10>\n" \
                    "\n" \
                    "    def test_failure(self):\n" \
                    ">       assert 0\n" \
                    "E       assert 0\n" \
                    "\n" \
                    "test_module.py:10: AssertionError\n"
    expected_cleaned_error = "self = <test_module.TestClass object at 0x7f4b2d2c1f10>\n" \
                             "\n" \
                             "    def test_failure(self):\n" \
                             ">       assert 0\n" \
                             "E       assert 0\n" \
                             "\n" \
                             "test_module.py:10: AssertionError\n"
    assert clean_error(error_message) == expected_cleaned_error

# Test to cover branch 177->182
def test_clean_error_with_summary_info():
    error_message = "self = <test_module.TestClass object at 0x7f4b2d2c1f10>\n" \
                    "\n" \
                    "    def test_failure(self):\n" \
                    ">       assert 0\n" \
                    "E       assert 0\n" \
                    "\n" \
                    "test_module.py:10: AssertionError\n" \
                    "=== short test summary info ====\n" \
                    "FAILED test_module.py::TestClass::test_failure - assert 0\n"
    expected_cleaned_error = "self = <test_module.TestClass object at 0x7f4b2d2c1f10>\n" \
                             "\n" \
                             "    def test_failure(self):\n" \
                             ">       assert 0\n" \
                             "E       assert 0\n" \
                             "\n" \
                             "test_module.py:10: AssertionError\n"
    assert clean_error(error_message) == expected_cleaned_error



import ast
import copy
from pathlib import Path
import typing as T
import importlib.util

_debug = lambda x: None


class Module(ast.Module):
    def __init__(self, original: ast.Module, path: Path):
        super().__init__(original.body, original.type_ignores)
        self.path = path

    def __reduce__(self):
        # for pickle/deepcopy
        return (self.__class__, (ast.Module(self.body, self.type_ignores), self.path))


# TODO use 'ast' alternative that retains comments?

def _package_path(file: Path) -> Path|None:
    """Returns a Python source file's path relative to its package"""
    import sys

    if not file.is_absolute():
        file = file.resolve()

    parents = list(file.parents)

    path: str|Path
    for path in sys.path:
        path = Path(path)
        if not path.is_absolute():
            path = path.resolve()

        for p in parents:
            if p == path:
                return file.relative_to(p)

    return None


def _get_fqn(file: Path) -> T.Sequence[str]|None:
    """Returns a source file's Python Fully Qualified Name, as a list name parts."""
    if not (path := _package_path(file)):
        return None

    path = path.parent if path.name == '__init__.py' else path.parent / path.stem
    return path.parts


def _resolve_from_import(file: Path, imp: ast.ImportFrom) -> str|None:
    """Resolves the module name in a `from X import Y` statement."""

    if imp.level > 0:  # relative import
        if not (pkg_path := _package_path(file)):
            return None

        pkg_path_parts = pkg_path.parts
        if imp.level > len(pkg_path_parts):
            return None # would go beyond top-level package

        return ".".join(pkg_path_parts[:-imp.level]) + (f".{imp.module}" if imp.module else "")

    return imp.module # absolute from ... import 


def _load_module(module_name: str) -> Module | None:
    try:
        if ((spec := importlib.util.find_spec(module_name))
            and spec.origin and spec.origin.endswith('.py')):
            _debug(f"Loading {spec.origin}")
            return parse_file(Path(spec.origin))

    except ModuleNotFoundError:
        pass

    return None


def _auto_stack(func):
    """Decorator that adds a stack of the first argument of the function being called."""
    def helper(*args):
        helper.stack.append(args[0])
        _debug(f"{'.'.join((n.name if getattr(n, 'name', None) else '<' + type(n).__name__ + '>') for n in helper.stack)}")
        retval = func(*args)
        helper.stack.pop()
        return retval
    helper.stack = []
    return helper


def _handle_import(
    module: Module,
    node: ast.AST,
    name: list[str],
    *,
    paths_seen: set[Path]|None = None
) -> list[Module|ast.stmt] | None:

    def transition(
        node: ast.Import | ast.ImportFrom,
        alias: ast.alias,
        mod: Module
    ) -> list[Module|ast.stmt]:
        imp = copy.copy(node)
        imp.names = [alias]
        return [imp, mod]

    if isinstance(node, ast.Import):
        # import N
        # import N.x                imports N and N.x
        # import a.b as N           'a.b' is renamed 'N'
        for alias in node.names:
            if alias.asname:
                if alias.asname == name[0]:
                    mod = _load_module(alias.name)
                    if path := _find_name_path(mod, name[1:], paths_seen=paths_seen):
                        # _find_name_path returns None if mod is None
                        return transition(node, alias, T.cast(Module, mod)) + path

            elif (import_name := alias.name.split('.'))[0] == name[0]:
                common_prefix = _common_prefix_len(import_name, name)
                mod = _load_module('.'.join(import_name[:common_prefix]))
                if path := _find_name_path(mod, name[common_prefix:], paths_seen=paths_seen):
                    return transition(node, alias, T.cast(Module, mod)) + path

    elif isinstance(node, ast.ImportFrom):
        # from a.b import N         either gets symbol N out of a.b, or imports a.b.N as N
        # from a.b import c as N

        for alias in node.names:
            if alias.name == '*':
                modname = _resolve_from_import(module.path, node)
                if (mod := _load_module(modname)) and \
                   (path := _find_name_path(mod, name, paths_seen=paths_seen)):
                    return transition(node, alias, mod) + path

            elif (alias.asname if alias.asname else alias.name) == name[0]:
                modname = _resolve_from_import(module.path, node)
                _debug(f"looking for symbol ({[alias.name, *name[1:]]} in {modname})")
                mod = _load_module(modname)
                if path := _find_name_path(mod, [alias.name, *name[1:]], paths_seen=paths_seen):
                    return transition(node, alias, T.cast(Module, mod)) + path

                _debug(f"looking for module ({name[1:]} in {modname}.{alias.name})")
                if (mod := _load_module(f"{modname}.{alias.name}")) and \
                   (path := _find_name_path(mod, name[1:], paths_seen=paths_seen)):
                    return transition(node, alias, T.cast(Module, mod)) + path

    return None


def _find_name_path(
    module: Module|None,
    name: list[str],
    *,
    paths_seen: set[Path]|None = None
) -> list[Module|ast.stmt]|None:
    """Looks for a symbol's definition by its name, returning the "path" of ast.ClassDef, ast.Import, etc.,
       crossed to find it.
    """
    if not module: return None
    if not name: return [module]

    _debug(f"looking up {name} in {module.path}")

    if not paths_seen: paths_seen = set()
    if module.path in paths_seen: return None
    paths_seen.add(module.path)

    @_auto_stack
    def find_name(node: ast.AST, name: list[str]) -> list[Module|ast.stmt]:
        _debug(f"_find_name {name} in {ast.dump(node)}")
        if isinstance(node, (ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef)):
            if node.name == name[0]:
                if len(name) == 1:
                    return [node]

                if isinstance(node, ast.ClassDef):
                    for stmt in node.body:
                        _debug(f"{node.name} checking {ast.dump(stmt)}")
                        if (path := find_name(stmt, name[1:])):
                            return [node, *path]

                    for base in node.bases:
                        base_name = ast.unparse(base).split('.')
                        if (len(find_name.stack) > 1 and
                            isinstance(context := find_name.stack[-2], ast.ClassDef)):
                            if (base_path := find_name(context, [context.name, *base_name, *name[1:]])):
                                return base_path[1:]

                        if (path := find_name(module, [*base_name, *name[1:]])):
                            return path

            return []

        if (isinstance(node, ast.Assign) and
            any(isinstance(n, ast.Name) and n.id == name[0] for t in node.targets for n in ast.walk(t))):
            return [node] if len(name) == 1 else []

        if isinstance(node, (ast.Import, ast.ImportFrom)):
            if (path := _handle_import(T.cast(Module, module), node, name, paths_seen=paths_seen)):
                return path

            return []

        # not isinstance(...) is just to trim down unnecessary work
        elif not isinstance(node, (ast.Expression, ast.Expr, ast.Name, ast.Attribute, ast.Compare)):
            for c in ast.iter_child_nodes(node):
                if (path := find_name(c, name)):
                    return path

        return []

    return find_name(module, name)


def _summarize(path: list[Module|ast.stmt]) -> Module|ast.stmt:
    """Replaces portions of Class elements with "..." to reduce noise and tokens."""

    # first the object
    if isinstance(path[-1], ast.ClassDef):
        path[-1] = copy.deepcopy(path[-1])
        for c in ast.iter_child_nodes(path[-1]):
            if (isinstance(c, ast.ClassDef) or
                (isinstance(c, (ast.FunctionDef, ast.AsyncFunctionDef)) and \
                 c.name not in ("__init__", "__new__"))):
                # Leave "__init__" unmodified as it's likely to contain important member information
                c.body = [ast.Expr(ast.Constant(value=ast.literal_eval("...")))]

    elif isinstance(path[-1], Module):
        path[-1] = copy.deepcopy(path[-1])
        for c in ast.iter_child_nodes(path[-1]):
            if isinstance(c, (ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef)):
                c.body = [ast.Expr(ast.Constant(value=ast.literal_eval("...")))]

    # now Class objects
    for i in reversed(range(len(path)-1)):
        if isinstance(path[i], ast.ClassDef):
            path[i] = copy.copy(path[i])
            path_i = T.cast(ast.ClassDef, path[i])
            path_i.body = [
                ast.Expr(ast.Constant(value=ast.literal_eval("..."))),
                T.cast(ast.stmt, path[i+1])
            ]

    return path[0]


def parse_file(file: Path) -> Module:
    """Reads a python source file, annotating it with its path/filename."""
    with file.open("r") as f:
        tree = ast.parse(f.read())

    return Module(tree, file)


def _common_prefix_len(a: T.Sequence[str], b: T.Sequence[str]) -> int:
    return next((i for i, (x, y) in enumerate(zip(a, b)) if x != y), min(len(a), len(b)))


def get_global_imports(module: Module, node: ast.AST) -> T.Sequence[ast.Import | ast.ImportFrom]:
    """Looks for module-level `import`s that (may) define the names seen in "node"."""

    def get_names(node: ast.AST):
        # TODO this ignores numerous ways in which a global import might not be visible,
        # such as when local symbols are created, etc.  In such cases, showing the
        # import in the excerpt is extraneous, but not incorrect.
        for n in ast.walk(node):
            if isinstance(n, ast.Name):
                yield n.id

    def get_imports(n: ast.AST):
        # TODO imports inside Class defines name in the class' namespace; they are uncommon
        # Imports within functions are included in the excerpt, so there's no need for us
        # to find them.
        if not isinstance(n, (ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef)):
            for child in ast.iter_child_nodes(n):
                yield from get_imports(child)

        if isinstance(n, (ast.Import, ast.ImportFrom)):
            yield n

    imports = []

    names = set(get_names(node))
    if names:
        for imp in get_imports(module):
            new_imp = None

            for alias in imp.names:
                alias_name = alias.asname if alias.asname else alias.name.split('.')[0]
                if alias_name in names:
                    if not new_imp:
                        new_imp = copy.copy(imp)
                        if isinstance(new_imp, ast.ImportFrom):
                            # make all imports absolute, as GPT, seeing relative imports in
                            # modules, may want to repeat them in tests as well
                            new_imp.module = _resolve_from_import(module.path, new_imp)
                            new_imp.level = 0

                        new_imp.names = []
                        imports.append(new_imp)
                    new_imp.names.append(alias)
                    names.remove(alias_name)

            if not names:
                break

    return imports


def _find_excerpt(node: ast.AST, line: int) -> ast.AST|None:
    if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
        begin = min([node.lineno] + [d.lineno for d in node.decorator_list])
        if begin <= line:
            return node

    for c in ast.iter_child_nodes(node):
        if (excerpt := _find_excerpt(c, line)):
            return excerpt

    return None


def get_info(module: Module, name: str, *, line: int = 0, generate_imports: bool = True) -> T.Optional[str]:
    """Returns summarized information on a class or function, following imports if necessary."""

    key = name.split('.')

    if (len(key) > 1 and (module_fqn := _get_fqn(module.path)) and
        (common_prefix := _common_prefix_len(module_fqn, key))):
        key = key[common_prefix:]

    path: list[Module|ast.stmt]|None = None
    # first look in the excerpt node, such as the focal function, if specified
    if (excerpt_node := _find_excerpt(module, line)):
        for c in ast.walk(excerpt_node):
            if (path := _handle_import(module, c, key)):
                break

    if not path:
        # Try looking among globals and classes
        path = _find_name_path(module, key)

    if not path:
        # Couldn't find the name in the context of the given module;
        # try to interpret it as an absolute FQN (GPT asks for that sometimes)
        key = name.split('.')
        for i in range(len(key)-1, 0, -1):
            if (mod := _load_module('.'.join(key[:i]))) and (path := _find_name_path(mod, key[i:])):
                break


    def any_import_as_or_import_in_class(path: T.Sequence[Module|ast.stmt]) -> bool:
        return any(
            isinstance(n, (ast.Import, ast.ImportFrom)) and (
                n.names[0].asname or (
                    i > 0 and isinstance(path[i-1], ast.ClassDef)
                )
            )
            for i, n in enumerate(path)
        )

    if path:
        _summarize(path)

        for i in range(len(path)):
            _debug(f"path[{i}]={ast.dump(path[i])}")

        if any_import_as_or_import_in_class(path):
            # include the full path for best context 
            path.insert(0, module)
        else:
            # just include the last module
            modules = [i for i in range(len(path)) if isinstance(path[i], ast.Module)]
            if modules:
                path = path[modules[-1]:]

        if not any(isinstance(n, ast.Module) for n in path):
            # no module in the path: we stay within the same file
            content = T.cast(ast.stmt, path[0])
            imports = get_global_imports(module, content) if generate_imports else []
            return f"""\
```python
{ast.unparse(ast.Module(body=[*imports, content], type_ignores=[]))}
```"""

        # there's a module in the path: build 'result' with all file transitions
        result = ""
        for i, item in enumerate(path):
            if isinstance(item, ast.Module):
                mod = item
                if i >= len(path)-1: # ends in a module
                    # When a module itself is the content, all imports are retained,
                    # so there's no need to look for them.
                    if result: result += "\n\n"
                    result += f"""\
in {_package_path(mod.path)}:
```python
{ast.unparse(mod)}
```"""
                else:
                    # continues to some content within a module
                    content = T.cast(ast.stmt, path[i+1])
                    imports = get_global_imports(mod, content) if generate_imports else []
                    if result: result += "\n\n"
                    result += f"""\
in {_package_path(mod.path)}:
```python
{ast.unparse(ast.Module(body=[*imports, content], type_ignores=[]))}
```"""
        return result

    return None



import abc
import argparse
import typing as T
from pathlib import Path
from ..utils import lines_branches_do
from ..segment import CodeSegment


class Prompter(abc.ABC):
    """Interface for a CoverUp prompter."""

    def __init__(self, cmd_args: argparse.Namespace):
        self.args = cmd_args


    @abc.abstractmethod
    def initial_prompt(self, segment: CodeSegment) -> T.List[dict]:
        """Returns initial prompt(s) for a code segment."""


    @abc.abstractmethod
    def error_prompt(self, segment: CodeSegment, error: str) -> T.List[dict] | None:
        """Returns prompts(s) in response to an error."""


    @abc.abstractmethod
    def missing_coverage_prompt(self, segment: CodeSegment,
                                missing_lines: set, missing_branches: set) -> T.List[dict] | None:
        """Returns prompts(s) in response to the suggested test(s) lacking coverage."""


    def get_functions(self) -> T.List[T.Callable]:
        """Returns a list of functions to be made available to the LLM.
           Each function's docstring must consist of its schema in JSON format."""
        return []


def get_module_name(src_file: Path, base_dir: Path) -> str|None:
    # assumes both src_file and src_dir Path.resolve()'d
    try:
        relative = src_file.relative_to(base_dir)
        return ".".join(relative.parts[:-1] + (relative.stem,))
    except ValueError:
        return None  # not relative to source


def mk_message(content: str, *, role="user") -> dict:
    return {
        'role': role,
        'content': content
    }



import typing as T
import openai
import logging
import asyncio
import warnings
import textwrap
import json
import traceback
from aiolimiter import AsyncLimiter

with warnings.catch_warnings():
    # ignore pydantic warnings https://github.com/BerriAI/litellm/issues/2832
    warnings.simplefilter('ignore')
    import litellm # type: ignore


# Turn off most logging
litellm.set_verbose = False
litellm.suppress_debug_info = True
logging.getLogger().setLevel(logging.ERROR)

# Ignore unavailable parameters
litellm.drop_params=True


# Tier 5 rate limits for models; tuples indicate limit and interval in seconds
# Extracted from https://platform.openai.com/account/limits on 8/30/2024
MODEL_RATE_LIMITS = {
    'gpt-3.5-turbo': {
        'token': (50_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-3.5-turbo-0125': {
        'token': (50_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-3.5-turbo-1106': {
        'token': (50_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-3.5-turbo-16k': {
        'token': (50_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-3.5-turbo-instruct': {
        'token': (90_000, 60), 'request': (3_500, 60)
    },
    'gpt-3.5-turbo-instruct-0914': {
        'token': (90_000, 60), 'request': (3_500, 60)
    },
    'gpt-4': {
        'token': (1_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-4-0314': {
        'token': (1_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-4-0613': {
        'token': (1_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-4-32k-0314': {
        'token': (150_000, 60), 'request': (1_000, 60)
    },
    'gpt-4-turbo': {
        'token': (2_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-4-turbo-2024-04-09': {
        'token': (2_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-4-turbo-preview': {
        'token': (2_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-4-0125-preview': {
        'token': (2_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-4-1106-preview': {
        'token': (2_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-4o': {
        'token': (30_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-4o-2024-05-13': {
        'token': (30_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-4o-2024-08-06': {
        'token': (30_000_000, 60), 'request': (10_000, 60)
    },
    'gpt-4o-mini': {
        'token': (150_000_000, 60), 'request': (30_000, 60)
    },
    'gpt-4o-mini-2024-07-18': {
        'token': (150_000_000, 60), 'request': (30_000, 60)
    }
}


def token_rate_limit_for_model(model_name: str) -> T.Tuple[int, int]|None:
    if model_name.startswith('openai/'):
        model_name = model_name[7:]

    if (model_limits := MODEL_RATE_LIMITS.get(model_name)):
        limit = model_limits.get('token')

        try:
            import tiktoken
            tiktoken.encoding_for_model(model_name)
        except KeyError:
            warnings.warn(f"Unable to get encoding for {model_name}; will ignore rate limit")
            return None

        return limit

    return None


def compute_cost(usage: dict, model_name: str) -> float|None:
    from math import ceil

    if model_name.startswith('openai/'):
        model_name = model_name[7:]

    if 'prompt_tokens' in usage and 'completion_tokens' in usage:
        if (cost := litellm.model_cost.get(model_name)):
            return usage['prompt_tokens'] * cost['input_cost_per_token'] +\
                   usage['completion_tokens'] * cost['output_cost_per_token']

    return None


_token_encoding_cache: dict[str, T.Any] = dict()
def count_tokens(model_name: str, completion: dict):
    """Counts the number of tokens in a chat completion request."""
    import tiktoken

    if not (encoding := _token_encoding_cache.get(model_name)):
        model = model_name
        if model_name.startswith('openai/'):
            model = model_name[7:]

        encoding = _token_encoding_cache[model_name] = tiktoken.encoding_for_model(model)

    count = 0
    for m in completion['messages']:
        count += len(encoding.encode(m['content']))

    return count


class ChatterError(Exception):
    pass


class Chatter:
    """Chats with an LLM."""

    def __init__(self, model: str) -> None:
        Chatter._validate_model(model)

        self._model = model
        self._model_temperature: float|None = None
        self._max_backoff = 64 # seconds
        self.token_rate_limit: AsyncLimiter|None
        self.set_token_rate_limit(token_rate_limit_for_model(model))
        self._add_cost = lambda cost: None
        self._log_msg = lambda ctx, msg: None
        self._log_json = lambda ctx, j: None
        self._signal_retry = lambda: None
        self._functions: dict[str, dict[str, T.Any]] = dict()
        self._max_func_calls_per_chat = 50


    @staticmethod
    def _validate_model(model) -> None:
        try:
            _, provider, _, _ = litellm.get_llm_provider(model)
        except litellm.exceptions.BadRequestError:
            raise ChatterError(textwrap.dedent("""\
                Unknown or unsupported model.
                Please see https://docs.litellm.ai/docs/providers for supported models."""
            ))

        result = litellm.validate_environment(model)
        if result['missing_keys']:
            if provider == 'openai':
                raise ChatterError(textwrap.dedent("""\
                    You need an OpenAI key to use {model}.
                    
                    You can get a key here: https://platform.openai.com/api-keys
                    Set the environment variable OPENAI_API_KEY to your key value
                        export OPENAI_API_KEY=<your key>"""
                ))
            elif provider == 'bedrock':
                raise ChatterError(textwrap.dedent("""\
                    To use Bedrock, you need an AWS account. Set the following environment variables:
                       export AWS_ACCESS_KEY_ID=<your key id>
                       export AWS_SECRET_ACCESS_KEY=<your secret key>
                       export AWS_REGION_NAME=us-west-2

                    You also need to request access to Claude:
                       https://docs.aws.amazon.com/bedrock/latest/userguide/model-access.html#manage-model-access"""
                ))
            else:
                raise ChatterError(textwrap.dedent(f"""\
                    You need a key (or keys) from {provider} to use {model}.
                    Set the following environment variables:
                        {', '.join(result['missing_keys'])}"""
                ))


    def set_model_temperature(self, temperature: T.Optional[float]) -> None:
        self._model_temperature = temperature


    def set_token_rate_limit(self, limit: T.Union[T.Tuple[int, int], None]) -> None:
        if limit:
            self.token_rate_limit = AsyncLimiter(*limit)
        else:
            self.token_rate_limit = None


    def set_max_backoff(self, max_backoff: int) -> None:
        self._max_backoff = max_backoff


    def set_add_cost(self, add_cost: T.Callable) -> None:
        """Sets up a callback to indicate additional costs."""
        self._add_cost = add_cost


    def set_log_msg(self, log_msg: T.Callable[[str, str], None]) -> None:
        """Sets up a callback to write a message to the log."""
        self._log_msg = log_msg


    def set_log_json(self, log_json: T.Callable[[str, dict], None]) -> None:
        """Sets up a callback to write a json exchange to the log."""
        self._log_json = log_json


    def set_signal_retry(self, signal_retry: T.Callable) -> None:
        """Sets up a callback to indicate a retry."""
        self._signal_retry = signal_retry


    def add_function(self, function: T.Callable) -> None:
        """Makes a function availabe to the LLM."""
        if not litellm.supports_function_calling(self._model):
            raise ChatterError(f"The {self._model} model does not support function calling.")

        try:
            schema = json.loads(getattr(function, "__doc__", ""))
            if 'name' not in schema:
                raise ChatterError("Name missing from function {function} schema.")
        except json.decoder.JSONDecodeError as e:
            raise ChatterError(f"Invalid JSON in function docstring: {e}")

        assert schema['name'] not in self._functions, "Duplicated function name {schema['name']}"
        self._functions[schema['name']] = {"function": function, "schema": schema}


    def _request(self, messages: T.List[dict]) -> dict:
        return {
            'model': self._model,
            **({'temperature': self._model_temperature} if self._model_temperature is not None else {}),
            'messages': messages,
            **({'api_base': "http://localhost:11434"} if "ollama" in self._model else {}),
            **({'tools': [{'type': 'function', 'function': f['schema']} for f in self._functions.values()]} \
                    if self._functions else {})
        }


    async def _send_request(self, request: dict, ctx: object) -> litellm.ModelResponse|None:
        """Sends the LLM chat request, handling common failures and returning the response."""

        sleep = 1
        while True:
            try:
                # TODO also add request limit; could use 'await asyncio.gather(t.acquire(tokens), r.acquire())'
                # to acquire both
                if self.token_rate_limit:
                    try:
                        await self.token_rate_limit.acquire(count_tokens(self._model, request))
                    except ValueError as e:
                        self._log_msg(ctx, f"Error: too many tokens for rate limit ({e})")
                        return None # gives up this segment

                return await litellm.acreate(**request)

            except (litellm.exceptions.ServiceUnavailableError,
                    openai.RateLimitError,
                    openai.APITimeoutError) as e:

                # This message usually indicates out of money in account
                if 'You exceeded your current quota' in str(e):
                    self._log_msg(ctx, f"Failed: {type(e)} {e}")
                    raise

                self._log_msg(ctx, f"Error: {type(e)} {e}")

                import random
                sleep = min(sleep*2, self._max_backoff)
                sleep_time = random.uniform(sleep/2, sleep)
                self._signal_retry()
                await asyncio.sleep(sleep_time)

            except openai.BadRequestError as e:
                # usually "maximum context length" XXX check for this?
                self._log_msg(ctx, f"Error: {type(e)} {e}")
                return None # gives up this segment

            except openai.AuthenticationError as e:
                self._log_msg(ctx, f"Failed: {type(e)} {e}")
                raise

            except openai.APIConnectionError as e:
                self._log_msg(ctx, f"Error: {type(e)} {e}")
                # usually a server-side error... just retry right away
                self._signal_retry()

            except openai.APIError as e:
                # APIError is the base class for all API errors;
                # we may be missing a more specific handler.
                print(f"Error: {type(e)} {e}; missing handler?")
                self._log_msg(ctx, f"Error: {type(e)} {e}")
                return None # gives up this segment


    def _call_function(self, ctx: object, tool_call: litellm.ModelResponse) -> str:
        args = json.loads(tool_call.function.arguments)
        function = self._functions[tool_call.function.name]

        try:
            return str(function['function'](ctx=ctx, **args))
        except Exception as e:
            self._log_msg(ctx, f"""\
Error executing function "{tool_call.function.name}": {e}
args:{args}

{traceback.format_exc()}
""")
            return f'Error executing function: {e}'


    async def chat(self, messages: list, *, ctx: T.Optional[object] = None) -> dict|None:
        """Chats with the LLM, sending the given messages, handling common failures and returning the response.
           Automatically calls any tool functions requested."""

        func_calls = 0
        while func_calls <= self._max_func_calls_per_chat:
            request = self._request(messages)
            self._log_json(ctx, request)

            if not (response := await self._send_request(request, ctx=ctx)):
                return None

            self._log_json(ctx, response.json())
            self._add_cost(litellm.completion_cost(response))

            if response.choices[0].finish_reason != "tool_calls":
                return response.json()

            tool_message = response.choices[0].message.json()
            tool_message["content"] = "" # it's typically set to null, which upsets count_tokens
            messages.append(tool_message)

            for call in response.choices[0].message.tool_calls:
                func_calls += 1
                messages.append({
                    'tool_call_id': call.id,
                    'role': 'tool',
                    'name': call.function.name,
                    'content': self._call_function(ctx, call)
                })

        self._log_msg(ctx, f"Too many function call requests, giving up")
        return None



import textwrap
import ast
from pathlib import Path
import typing as T
import pytest
import coverup.codeinfo as codeinfo
import sys


codeinfo._debug = print     # enables debugging


@pytest.fixture
def importlib_cleanup():
    import importlib

    previously_loaded = {m for m in sys.modules}

    yield

    importlib.invalidate_caches()

    # It's not enough to call 'invalidate_caches'... :/
    for m in list(sys.modules):
        if m not in previously_loaded:
            del sys.modules[m]


@pytest.fixture
def import_fixture(importlib_cleanup, monkeypatch):
    import tempfile
    import shutil

    # Avoid using the tmp_path fixture because it retains the directories
    # while other tests execute... if importlib then still has something
    # cached, we confusingly get other tests' contents, rather than clean
    # "file not found" errors.
    with tempfile.TemporaryDirectory(dir=Path('.')) as tmpdir:
        tmp_path = Path(tmpdir).resolve()
        monkeypatch.chdir(tmp_path)
        monkeypatch.syspath_prepend(tmp_path)
        yield tmp_path

    shutil.rmtree(tmp_path, ignore_errors=True)


def get_fqn(p):
    fqn = codeinfo._get_fqn(p)
    return '.'.join(fqn) if fqn else fqn


def test_get_fqn(import_fixture):
    tmp_path = import_fixture

    (tmp_path / "foo").mkdir()

    assert "foo" == get_fqn(tmp_path / "foo" / "__init__.py")
    assert "foo.bar" == get_fqn(tmp_path / "foo" / "bar.py")
    assert "foo.bar.baz" == get_fqn(tmp_path / "foo" / "bar" / "baz.py")

    assert "foo" == get_fqn(Path("foo") / "__init__.py")
    assert "foo.bar" == get_fqn(Path("foo") / "bar.py")
    assert "foo.bar.baz" == get_fqn(Path("foo") / "bar" / "baz.py")


def test_get_fqn_relative_syspath(importlib_cleanup, monkeypatch):
    import tempfile

    with tempfile.TemporaryDirectory(dir=Path('.')) as tmpdir:
        tmp_path = Path(tmpdir).resolve()
        rel_path = tmp_path.relative_to(Path.cwd())

        monkeypatch.chdir(rel_path)
        monkeypatch.syspath_prepend('.')

        assert "foo" == get_fqn(tmp_path / "foo" / "__init__.py")
        assert "foo.bar" == get_fqn(tmp_path / "foo" / "bar.py")
        assert "foo.bar.baz" == get_fqn(tmp_path / "foo" / "bar" / "baz.py")

        assert "foo" == get_fqn(Path("foo") / "__init__.py")
        assert "foo.bar" == get_fqn(Path("foo") / "bar.py")
        assert "foo.bar.baz" == get_fqn(Path("foo") / "bar" / "baz.py")


def test_resolve_from_import(import_fixture):
    tmp_path = import_fixture

    def I(imp_code):
        return ast.parse(imp_code).body[0]

    (tmp_path / "code.py").touch()
    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").touch()
    (tmp_path / "foo" / "bar").mkdir()
    (tmp_path / "foo" / "bar" / "__init__.py").touch()
    (tmp_path / "foo" / "bar" / "none.py").touch()
    (tmp_path / "foo" / "baz.py").touch()
    (tmp_path / "bar.py").touch()

    foo = tmp_path / "foo"
    file = foo / "__init__.py"
    assert "foo" == codeinfo._resolve_from_import(file, I("from . import bar"))
    assert "foo.bar" == codeinfo._resolve_from_import(file, I("from .bar import baz"))
    assert "foo.bar.none" == codeinfo._resolve_from_import(file, I("from .bar.none import baz"))

    file = foo / "bar" / "__init__.py"
    assert "foo.bar" == codeinfo._resolve_from_import(file, I("from . import none"))
    assert "foo.bar.none" == codeinfo._resolve_from_import(file, I("from .none import *"))
    assert "foo.baz" == codeinfo._resolve_from_import(file, I("from ..baz import *"))
    assert "foo.bar.none" == codeinfo._resolve_from_import(file, I("from ..bar.none import *"))


def test_get_info_class(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "foo.py"
    code.write_text(textwrap.dedent("""\
        class C:
            x = 10

            def __init__(self, x: int) -> C:
                self._foo = x

            class B:
                z = 42

                def foo():
                    '''far niente'''
                    pass

            @deco
            def foo(self) -> None:
                print(self._foo)

            def bar(self):
                pass

        def func(x: C):
            x.foo()

        class D:
            def __new__(cls, *args, **kwargs):
                return super().__new__(cls)

            def d(self):
                pass
        """
    ))

    tree = codeinfo.parse_file(code)

    assert codeinfo.get_info(tree, 'C') == textwrap.dedent("""\
        ```python
        class C:
            x = 10

            def __init__(self, x: int) -> C:
                self._foo = x

            class B:
                ...

            @deco
            def foo(self) -> None:
                ...

            def bar(self):
                ...
        ```"""
    )

    assert codeinfo.get_info(tree, 'D') == textwrap.dedent('''\
        ```python
        class D:

            def __new__(cls, *args, **kwargs):
                return super().__new__(cls)

            def d(self):
                ...
        ```'''
    )

    assert codeinfo.get_info(tree, 'C.foo') == textwrap.dedent("""\
        ```python
        class C:
            ...

            @deco
            def foo(self) -> None:
                print(self._foo)
        ```"""
    )

    assert codeinfo.get_info(tree, 'C.B') == textwrap.dedent("""\
        ```python
        class C:
            ...

            class B:
                z = 42

                def foo():
                    ...
        ```"""
    )

    assert codeinfo.get_info(tree, 'C.B.foo') == textwrap.dedent('''\
        ```python
        class C:
            ...

            class B:
                ...

                def foo():
                    """far niente"""
                    pass
        ```'''
    )

    assert codeinfo.get_info(tree, 'C.B.bar') == None
    assert codeinfo.get_info(tree, 'foo') == None


def test_get_info_method_from_parent(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "foo.py"
    code.write_text(textwrap.dedent("""\
        class A:
            def a(self):
                return 42

        class B(A):
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'B.a') == textwrap.dedent("""\
        ```python
        class A:
            ...

            def a(self):
                return 42
        ```"""
    )


def test_get_info_method_from_parent_is_attribute(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "foo.py"
    code.write_text(textwrap.dedent("""\
        class A:
            class B:
                def a(self):
                    return 42

        class C(A.B):
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'C.a') == textwrap.dedent("""\
        ```python
        class A:
            ...

            class B:
                ...

                def a(self):
                    return 42
        ```"""
    )


def test_get_info_method_from_parent_parent_missing(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "foo.py"
    code.write_text(textwrap.dedent("""\
        from doesnotexist import A

        class B(A):
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'B.a') == None


def test_get_info_method_from_parent_imported(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "foo.py"
    code.write_text(textwrap.dedent("""\
        from bar import A

        class B(A):
            pass
        """
    ))

    (tmp_path / "bar.py").write_text(textwrap.dedent("""\
        class A:
            def a(self):
                return 42
        """
    ))

    # TODO it would be better if it showed class B(A)
    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'B.a') == textwrap.dedent("""\
        in bar.py:
        ```python
        class A:
            ...

            def a(self):
                return 42
        ```"""
    )


def test_get_info_method_from_parent_imported_in_class(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "foo.py"
    code.write_text(textwrap.dedent("""\
        class A:
            from bar import B

            class C(B):
                pass

            class D(C):
                pass
        """
    ))

    (tmp_path / "bar.py").write_text(textwrap.dedent("""\
        class B:
            def b(self):
                return 42
        """
    ))

    # FIXME this would be better if it showed class C(B) and class D(C)
    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'A.D.b') == textwrap.dedent("""\
        in foo.py:
        ```python
        class A:
            ...
            from bar import B
        ```

        in bar.py:
        ```python
        class B:
            ...

            def b(self):
                return 42
        ```"""
    )


def test_get_info_assignment(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "foo.py"
    code.write_text(textwrap.dedent("""\
        PI = 3.1415

        x = 0
        if x == 0:
            TYPES = (
                int,
                str,
            )

        class C:
            x = 10

            def __init__(self, x: int) -> C:
                self._foo = x
        """
    ))

    tree = codeinfo.parse_file(code)

    assert codeinfo.get_info(tree, 'PI') == textwrap.dedent("""\
        ```python
        PI = 3.1415
        ```"""
    )

    # FIXME do proper code slicing
    assert codeinfo.get_info(tree, 'TYPES') == textwrap.dedent("""\
        ```python
        TYPES = (int, str)
        ```"""
    )

    assert codeinfo.get_info(tree, 'C.x') == textwrap.dedent("""\
        ```python
        class C:
            ...
            x = 10
        ```"""
    )


def test_get_info_import(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import os
        import foo
        """
    ))

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text(textwrap.dedent("""\
        class Foo:
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'foo.Foo') == textwrap.dedent('''\
        in foo/__init__.py:
        ```python
        class Foo:
            pass
        ```'''
    )

    assert codeinfo.get_info(tree, 'foo.baz') == None # doesn't exist


def test_get_info_import_submodule(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import os
        import foo.bar
        """
    ))

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text(textwrap.dedent("""\
        class Foo:
            pass
        """
    ))
    (tmp_path / "foo" / "bar.py").write_text(textwrap.dedent("""\
        class Bar:
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'foo.Foo') == textwrap.dedent('''\
        in foo/__init__.py:
        ```python
        class Foo:
            pass
        ```'''
    )

    assert codeinfo.get_info(tree, 'foo.bar.Bar') == textwrap.dedent('''\
        in foo/bar.py:
        ```python
        class Bar:
            pass
        ```'''
    )


def test_get_info_import_as(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import os
        import foo as baz
        """
    ))

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text(textwrap.dedent("""\
        class Foo:
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'baz.Foo') == textwrap.dedent('''\
        in code.py:
        ```python
        import foo as baz
        ```

        in foo/__init__.py:
        ```python
        class Foo:
            pass
        ```'''
    )


def test_get_info_from_import_symbol_exists(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import os
        from foo import bar
        """
    ))

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text(textwrap.dedent("""\
        class Foo:
            pass

        class bar:  # 'bar' is defined here -- takes precedence over bar.py
            class Bar:
                pass
        """
    ))
    (tmp_path / "foo" / "bar.py").write_text(textwrap.dedent("""\
        class Bar:
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'bar.Bar') == textwrap.dedent('''\
        in foo/__init__.py:
        ```python
        class bar:
            ...

            class Bar:
                pass
        ```'''
    )


def test_get_info_from_import_as_symbol_exists(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import os
        from foo import bar as baz
        """
    ))

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text(textwrap.dedent("""\
        class Foo:
            pass

        class bar:  # 'bar' is defined here -- takes precedence over bar.py
            class Bar:
                pass
        """
    ))
    (tmp_path / "foo" / "bar.py").write_text(textwrap.dedent("""\
        class Bar:
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'baz.Bar') == textwrap.dedent('''\
        in code.py:
        ```python
        from foo import bar as baz
        ```

        in foo/__init__.py:
        ```python
        class bar:
            ...

            class Bar:
                pass
        ```'''
    )


def test_get_info_from_import_symbol_doesnt_exist(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import os
        from foo import bar
        """
    ))

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text(textwrap.dedent("""\
        class Foo:
            pass
        """
    ))
    (tmp_path / "foo" / "bar.py").write_text(textwrap.dedent("""\
        class Bar:
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'bar.Bar') == textwrap.dedent('''\
        in foo/bar.py:
        ```python
        class Bar:
            pass
        ```'''
    )


def test_get_info_from_import_as_symbol_doesnt_exist(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import os
        from foo import bar as baz
        """
    ))

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text(textwrap.dedent("""\
        class Foo:
            pass
        """
    ))
    (tmp_path / "foo" / "bar.py").write_text(textwrap.dedent("""\
        class Bar:
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'baz.Bar') == textwrap.dedent('''\
        in code.py:
        ```python
        from foo import bar as baz
        ```

        in foo/bar.py:
        ```python
        class Bar:
            pass
        ```'''
    )


def test_get_info_from_import_relative(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import os
        import foo
        """
    ))

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text(textwrap.dedent("""\
        from . import bar
        """
    ))
    (tmp_path / "foo" / "bar.py").write_text(textwrap.dedent("""\
        from .baz import Baz as Bar
        """
    ))
    (tmp_path / "foo" / "baz.py").write_text(textwrap.dedent("""\
        class Baz:
            answer = 42
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'foo.bar.Bar') == textwrap.dedent('''\
        in code.py:
        ```python
        import foo
        ```

        in foo/__init__.py:
        ```python
        from . import bar
        ```

        in foo/bar.py:
        ```python
        from .baz import Baz as Bar
        ```

        in foo/baz.py:
        ```python
        class Baz:
            answer = 42
        ```'''
    )


def test_get_info_import_in_excerpt_function(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import os

        def something():
            def something_else():
                from foo import Foo
                Foo()
        """
    ))

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text(textwrap.dedent("""\
        class Foo:
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'Foo', line=3) == textwrap.dedent('''\
        in foo/__init__.py:
        ```python
        class Foo:
            pass
        ```'''
    )


def test_get_info_import_in_excerpt_class(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import os

        class X:
            from foo import Foo

            def __init__(self, x: Foo):
                self.x = x
        """
    ))

    (tmp_path / "foo.py").write_text(textwrap.dedent("""\
        class Foo:
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'Foo', line=3) == textwrap.dedent('''\
        in foo.py:
        ```python
        class Foo:
            pass
        ```'''
    )


def test_get_info_import_in_class(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        class C:
            import foo
        """
    ))

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text(textwrap.dedent("""\
        from . import bar

        class Foo:
            pass
        """
    ))
    (tmp_path / "foo" / "bar.py").write_text(textwrap.dedent("""\
        class Bar:
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'C.foo.bar.Bar') == textwrap.dedent('''\
        in code.py:
        ```python
        class C:
            ...
            import foo
        ```

        in foo/__init__.py:
        ```python
        from . import bar
        ```

        in foo/bar.py:
        ```python
        class Bar:
            pass
        ```'''
    )


def test_get_info_imported_assignment(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import foo.constants as C
        """
    ))

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text("")
    (tmp_path / "foo" / "constants.py").write_text(textwrap.dedent("""\
        PI = 3.1415926
        """
    ))

    tree = codeinfo.parse_file(code)

    assert codeinfo.get_info(tree, 'C.PI') == textwrap.dedent('''\
        in code.py:
        ```python
        import foo.constants as C
        ```

        in foo/constants.py:
        ```python
        PI = 3.1415926
        ```'''
    )


def test_get_info_import_and_class_in_block(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        try:
            import foo
        except ImportError:
            import bar as foo
        """
    ))

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text(textwrap.dedent("""\
        if True:
            class Foo:
                pass
        """
    ))
    (tmp_path / "bar.py").write_text(textwrap.dedent("""\
        class Bar:
            answer = 42
        """
    ))

    tree = codeinfo.parse_file(code)

    assert codeinfo.get_info(tree, 'foo.Foo') == textwrap.dedent('''\
        in foo/__init__.py:
        ```python
        class Foo:
            pass
        ```'''
    )


@pytest.mark.parametrize("from_module", [True, False])
def test_get_info_name_includes_module_fqn(import_fixture, from_module):
    tmp_path = import_fixture

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text("")
    (tmp_path / "foo" / "bar.py").write_text(textwrap.dedent("""\
        class C:
            pass
        """
    ))

    if from_module:
        tree = codeinfo.parse_file(tmp_path / "foo" / "bar.py")
        assert codeinfo.get_info(tree, 'foo.bar.C') == textwrap.dedent('''\
            ```python
            class C:
                pass
            ```'''
        )

    else:
        (tmp_path / "x.py").write_text(textwrap.dedent("""\
            from foo.bar import C
            """
        ))

        # class C really apears under the name "C", not "foo.bar.C", but gpt-4o sometimes
        # asks for names like it after seeing the equivalent of  "from foo.bar import C"
        tree = codeinfo.parse_file(tmp_path / "x.py")
        assert codeinfo.get_info(tree, 'foo.bar.C') == textwrap.dedent('''\
            ```python
            class C:
                pass
            ```'''
        )


def test_get_info_includes_imports(import_fixture):
    tmp_path = import_fixture
    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import os
        from foo import bar as R, baz as Z
        import sys, ast

        class C(Z):
            from foo import Foo
            x = len(sys.path)

            def __init__(self, x: int) -> C:
                self._foo = Z(x)

            class B:
                z = 42

        def func(x: C):
            x.foo(os.environ)
        """
    ))

    (tmp_path / "foo.py").write_text(textwrap.dedent("""\
        import sys
        from .bar import Bar

        class Foo(Bar):
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'C') == textwrap.dedent("""\
        ```python
        from foo import baz as Z
        import sys

        class C(Z):
            from foo import Foo
            x = len(sys.path)

            def __init__(self, x: int) -> C:
                self._foo = Z(x)

            class B:
                ...
        ```"""
    )

    assert codeinfo.get_info(tree, 'C', generate_imports=False) == textwrap.dedent("""\
        ```python
        class C(Z):
            from foo import Foo
            x = len(sys.path)

            def __init__(self, x: int) -> C:
                self._foo = Z(x)

            class B:
                ...
        ```"""
    )

    assert codeinfo.get_info(tree, 'C.Foo') == textwrap.dedent("""\
        in code.py:
        ```python
        from foo import baz as Z

        class C(Z):
            ...
            from foo import Foo
        ```

        in foo.py:
        ```python
        from .bar import Bar

        class Foo(Bar):
            pass
        ```"""
    )

    assert codeinfo.get_info(tree, 'C.Foo', generate_imports=False) == textwrap.dedent("""\
        in code.py:
        ```python
        class C(Z):
            ...
            from foo import Foo
        ```

        in foo.py:
        ```python
        class Foo(Bar):
            pass
        ```"""
    )

    assert codeinfo.get_info(tree, 'func') == textwrap.dedent("""\
        ```python
        import os

        def func(x: C):
            x.foo(os.environ)
        ```"""
    )

    assert codeinfo.get_info(tree, 'func', generate_imports=False) == textwrap.dedent("""\
        ```python
        def func(x: C):
            x.foo(os.environ)
        ```"""
    )


def test_get_info_included_imports_always_absolute(import_fixture):
    tmp_path = import_fixture
    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import foo
        """
    ))

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text(textwrap.dedent("""\
        import sys
        from .bar import Bar

        class Foo(Bar):
            pass
        """
    ))

    (tmp_path / "foo" / "bar.py").write_text(textwrap.dedent("""\
        class Bar:
            pass
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'foo.Foo') == textwrap.dedent("""\
        in foo/__init__.py:
        ```python
        from foo.bar import Bar

        class Foo(Bar):
            pass
        ```"""
    )


def test_get_global_imports(import_fixture):
    tmp_path = import_fixture
    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import a, b
        from c import d as e
        import os

        try:
            from hashlib import sha1
        except ImportError:
            from sha import sha as sha1
        
        class Foo:
            import abc as os

            def f():
                if os.path.exists("foo"):
                    sha1(a.x, b.x, e.x)
        """))

    print(code)

    def find_node(tree, name):
        for n in ast.walk(tree):
            if isinstance(n, (ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef)) and n.name == name:
                return n

    tree = codeinfo.parse_file(code)
#    print(ast.dump(tree, indent=2))

    f = find_node(tree, 'f')
    assert [ast.unparse(x) for x in codeinfo.get_global_imports(tree, f)] == [
        'import a, b',
        'from c import d as e',
        'import os',
        'from hashlib import sha1'
    ]


def test_get_info_module(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "foo.py"
    code.write_text(textwrap.dedent("""\
        import bar
        """
    ))
    (tmp_path / "bar.py").write_text(textwrap.dedent("""\
        try:
            import os
        except ImportError:
            import os2

        class A:
            def __init__(self):
                pass

            class B:
                pass

        def foo(a, b):
            return a+b

        answer = 42

        if __name__ == '__main__':
            import sys
            sys.exit(0)
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'bar') == textwrap.dedent("""\
        in bar.py:
        ```python
        try:
            import os
        except ImportError:
            import os2

        class A:
            ...

        def foo(a, b):
            ...
        answer = 42
        if __name__ == '__main__':
            import sys
            sys.exit(0)
        ```"""
    )


@pytest.mark.skipif(sys.version_info[0:2] < (3,11), reason="not a frozen module then")
def test_get_info_frozen_module(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "foo.py"
    code.write_text(textwrap.dedent("""\
        import os
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'os.path.join') == None


def test_get_info_builtin_module(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "foo.py"
    code.write_text(textwrap.dedent("""\
        import sys
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'sys.path') == None


def test_get_info_non_py(import_fixture):
    """On Python 3.10 on Linux, this module maps to a .so, not a Python file"""
    tmp_path = import_fixture

    code = tmp_path / "foo.py"
    code.write_text("")

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'termios.tcgetattr') == None


def test_get_info_import_star(import_fixture):
    tmp_path = import_fixture

    code = tmp_path / "code.py"
    code.write_text(textwrap.dedent("""\
        import foo
        """
    ))

    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "__init__.py").write_text(textwrap.dedent("""\
        import sys

        if sys.platform == 'win32':
            from .win32 import *
        else:
            from .unix import *
        """
    ))
    (tmp_path / "foo" / "unix.py").write_text(textwrap.dedent("""\
        answer = 'eunuchs'
        """
    ))
    (tmp_path / "foo" / "win32.py").write_text(textwrap.dedent("""\
        answer = '$$$$'
        """
    ))

    tree = codeinfo.parse_file(code)
    assert codeinfo.get_info(tree, 'foo.answer') == textwrap.dedent("""\
        in foo/win32.py:
        ```python
        answer = '$$$$'
        ```"""
    )



import typing as T
from .prompter import Prompter, CodeSegment, mk_message, get_module_name
from ..utils import lines_branches_do


class GptV1Prompter(Prompter):
    """Prompter for GPT-4."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def initial_prompt(self, segment: CodeSegment) -> T.List[dict]:
        module_name = get_module_name(segment.path, self.args.src_base_dir)
        filename = segment.path.relative_to(self.args.src_base_dir)

        return [
            mk_message(f"""
You are an expert Python test-driven developer.
The code below, extracted from {filename},{' module ' + module_name + ',' if module_name else ''} does not achieve full coverage:
when tested, {'it does' if not segment.executed_lines else segment.lines_branches_missing_do()} not execute.
Create a new pytest test function that executes these missing lines/branches, always making
sure that the new test is correct and indeed improves coverage.
Always send entire Python test scripts when proposing a new test or correcting one you
previously proposed.
Be sure to include assertions in the test that verify any applicable postconditions.
Please also make VERY SURE to clean up after the test, so as not to affect other tests;
use 'pytest-mock' if appropriate.
Write as little top-level code as possible, and in particular do not include any top-level code
calling into pytest.main or the test itself.
Respond ONLY with the Python code enclosed in backticks, without any explanation.
```python
{segment.get_excerpt(tag_lines=bool(segment.executed_lines))}
```
""")
        ]

    def error_prompt(self, segment: CodeSegment, error: str) -> T.List[dict]:
        return [mk_message(f"""\
Executing the test yields an error, shown below.
Modify the test to correct it; respond only with the complete Python code in backticks.

{error}""")
        ]


    def missing_coverage_prompt(self, segment: CodeSegment,
                                missing_lines: set, missing_branches: set) -> T.List[dict]:
        return [mk_message(f"""\
This test still lacks coverage: {lines_branches_do(missing_lines, set(), missing_branches)} not execute.
Modify it to correct that; respond only with the complete Python code in backticks.
""")
        ]



# file src/coverup/coverup.py:357-359
# lines [359]
# branches []

import pytest
from coverup.coverup import State

# Assuming the State class has an __init__ method that requires an initial_coverage argument.
# Since the code for the State class is not provided, we will mock the coverage attribute.

def test_get_initial_coverage():
    expected_coverage = {'file1.py': {'lines': [1, 2, 3]}, 'file2.py': {'lines': [4, 5, 6]}}
    state = State(initial_coverage=expected_coverage)
    
    # Call the method we want to test
    coverage = state.get_initial_coverage()
    
    # Assert that the method returns the correct coverage
    assert coverage == expected_coverage, "The get_initial_coverage method did not return the expected coverage data"



# file src/coverup/testrunner.py:40-63
# lines [62, 63]
# branches ['53->53']

import json
import os
import pytest
import subprocess
import sys
from pathlib import Path
from unittest.mock import patch

# Test to cover lines 62-63
def test_measure_suite_coverage_file_not_found(tmp_path, monkeypatch):
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    tests_dir = tmp_path / "tests"
    tests_dir.mkdir()
    test_file = tests_dir / "test_example.py"
    test_file.write_text("def test_pass(): assert True")

    def mock_run(*args, **kwargs):
        # Simulate the behavior of subprocess.run
        class MockCompletedProcess:
            def __init__(self):
                self.returncode = 0
                self.stdout = b""

            def check_returncode(self):
                pass

        return MockCompletedProcess()

    monkeypatch.setattr(subprocess, "run", mock_run)

    def mock_unlink(path):
        # Simulate the behavior of os.unlink when file is not found
        raise FileNotFoundError

    monkeypatch.setattr(os, "unlink", mock_unlink)

    def mock_json_load(file):
        # Simulate json.load returning an empty dictionary
        return {}

    monkeypatch.setattr(json, "load", mock_json_load)

    from coverup.testrunner import measure_suite_coverage

    # Run the test suite coverage measurement
    coverage_data = measure_suite_coverage(tests_dir=tests_dir, source_dir=source_dir)
    assert isinstance(coverage_data, dict)  # Postcondition: coverage_data should be a dict

# Test to cover branch 53->53
def test_measure_suite_coverage_non_zero_returncode(tmp_path, monkeypatch):
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    tests_dir = tmp_path / "tests"
    tests_dir.mkdir()
    test_file = tests_dir / "test_example.py"
    test_file.write_text("def test_fail(): assert False")

    def mock_run(*args, **kwargs):
        # Simulate the behavior of subprocess.run with non-zero returncode
        class MockCompletedProcess:
            def __init__(self):
                self.returncode = 1  # Non-zero returncode
                self.stdout = b""

            def check_returncode(self):
                raise subprocess.CalledProcessError(self.returncode, args)

        return MockCompletedProcess()

    monkeypatch.setattr(subprocess, "run", mock_run)

    from coverup.testrunner import measure_suite_coverage

    # Run the test suite coverage measurement and expect a CalledProcessError
    with pytest.raises(subprocess.CalledProcessError):
        measure_suite_coverage(tests_dir=tests_dir, source_dir=source_dir)

    # No postcondition to assert since the test is expected to raise an exception



# file src/coverup/utils.py:26-46
# lines [30, 31, 33, 34, 35, 36, 37, 39, 40, 42, 44, 46]
# branches ['34->exit', '34->35', '36->37', '36->39', '39->40', '39->42']

import pytest
from coverup.utils import format_ranges

def test_format_ranges_with_negative_ranges():
    # Define the sets of lines and negative lines to pass to the function
    lines = {1, 2, 4, 5}
    negative = {3}

    # Call the function with the test data
    result = format_ranges(lines, negative)

    # Verify the result is as expected
    assert result == "1-2, 4-5", "The format_ranges function did not return the expected string"



# file src/coverup/coverup.py:203-222
# lines [204, 206, 207, 208, 209, 211, 213, 214, 215, 216, 217, 219, 220, 222]
# branches ['213->214', '213->222', '214->215', '214->219', '215->213', '215->216', '216->215', '216->217', '219->213', '219->220']

import ast
import pytest
from coverup.coverup import find_imports

@pytest.fixture
def python_code_with_syntax_error():
    return "invalid python code"

@pytest.fixture
def python_code_with_imports():
    return """
import os
import sys
from pathlib import Path
import re
"""

def test_find_imports_with_syntax_error(python_code_with_syntax_error):
    result = find_imports(python_code_with_syntax_error)
    assert result == []

def test_find_imports_with_valid_code(python_code_with_imports):
    result = find_imports(python_code_with_imports)
    assert set(result) == {"os", "sys", "pathlib", "re"}



# file src/coverup/segment.py:67-68
# lines [68]
# branches []

import pytest
from coverup.segment import CodeSegment

# Assuming the CodeSegment class requires 9 arguments for initialization
# and has attributes `missing_lines` and `missing_branches`.

# Test to cover line 68
def test_missing_count():
    segment = CodeSegment(
        filename='test_file.py',
        name='test_segment',
        begin=1,
        end=10,
        lines_of_interest=set(range(1, 11)),
        missing_lines={1, 2, 3},
        executed_lines=set(range(4, 11)),
        missing_branches={(5, True), (6, False)},
        context=[], imports=[]
    )
    assert segment.missing_count() == 5

# Test to cover line 68 with no missing lines or branches
def test_missing_count_zero():
    segment = CodeSegment(
        filename='test_file.py',
        name='test_segment',
        begin=1,
        end=10,
        lines_of_interest=set(range(1, 11)),
        missing_lines=set(),
        executed_lines=set(range(1, 11)),
        missing_branches=set(),
        context=[], imports=[]
    )
    assert segment.missing_count() == 0

# Test to cover line 68 with missing lines but no missing branches
def test_missing_count_only_lines():
    segment = CodeSegment(
        filename='test_file.py',
        name='test_segment',
        begin=1,
        end=10,
        lines_of_interest=set(range(1, 11)),
        missing_lines={1, 2, 3},
        executed_lines=set(range(4, 11)),
        missing_branches=set(),
        context=[], imports=[]
    )
    assert segment.missing_count() == 3

# Test to cover line 68 with missing branches but no missing lines
def test_missing_count_only_branches():
    segment = CodeSegment(
        filename='test_file.py',
        name='test_segment',
        begin=1,
        end=10,
        lines_of_interest=set(range(1, 11)),
        missing_lines=set(),
        executed_lines=set(range(1, 11)),
        missing_branches={(5, True), (6, False)},
        context=[], imports=[]
    )
    assert segment.missing_count() == 2



import pytest
import subprocess
import coverup.testrunner as tr
from pathlib import Path
import tempfile
import textwrap
import sys


@pytest.mark.asyncio
async def test_measure_test_coverage_exit_1(tmpdir):
    with pytest.raises(subprocess.CalledProcessError) as einfo:
        await tr.measure_test_coverage(test="import os;\ndef test_foo(): os.exit(1)\n", tests_dir=Path(tmpdir))


@pytest.mark.skipif(sys.version_info[:2] == (3,10), reason="Python 3.10 doesn't use state() as newer versions do")
@pytest.mark.asyncio
async def test_measure_test_coverage_uses_cleanslate(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)

    (tmp_path / "foo.py").write_text(textwrap.dedent("""\
        answer = 42
        """
    ))
    test = textwrap.dedent("""\
        import os

        def test_one():
            from foo import answer
            assert answer == 42

        def test_two(monkeypatch):
            def mock_stat(path):
                if path == '/':
                    return os.stat_result((0, 0, 0, 0, 0, 0, 0, 0, 0, 0))
                raise Exception('mocked stat failure')

            monkeypatch.setattr(os, 'stat', mock_stat)
            from foo import answer      # should fail because of 'stat' if run in isolation
            assert answer == 42
        """)

    with pytest.raises(subprocess.CalledProcessError) as einfo:
        await tr.measure_test_coverage(test=test, tests_dir=tmp_path, isolate_tests=True)


@pytest.mark.parametrize("absolute", [True, False])
def test_measure_suite_coverage_empty_dir(absolute):
    with tempfile.TemporaryDirectory(dir=Path('.')) as tests_dir:
        tests_dir = Path(tests_dir)
        if absolute:
            tests_dir = tests_dir.resolve()

        coverage = tr.measure_suite_coverage(tests_dir=tests_dir, source_dir=tests_dir, trace=print)   # shouldn't throw
        assert {} == coverage['files']


def seq2p(tests_dir, seq):
    return tests_dir / f"test_coverup_{seq}.py"


N_TESTS=10
def make_failing_suite(tests_dir: Path, fail_collect: bool):
    """In a suite with 10 tests, test 6 fails; test 3 doesn't fail, but causes 6 to fail."""

    for seq in range(N_TESTS):
        seq2p(tests_dir, seq).write_text('def test_foo(): pass')

    culprit = seq2p(tests_dir, 3)
    culprit.write_text("import sys\n" + "sys.foobar = True\n" + "def test_foo(): pass")

    failing = seq2p(tests_dir, 6)
    if fail_collect:
        failing.write_text("import sys\n" + "assert not getattr(sys, 'foobar', False)\n" + "def test_foo(): pass")
    else:
        failing.write_text("import sys\n" + "def test_foo(): assert not getattr(sys, 'foobar', False)")

    return failing, culprit


@pytest.mark.parametrize("fail_collect", [True, False])
@pytest.mark.parametrize("absolute", [True, False])
def test_measure_suite_coverage_test_fails(absolute, fail_collect):
    with tempfile.TemporaryDirectory(dir=Path('.')) as tests_dir:
        tests_dir = Path(tests_dir)
        if absolute:
            tests_dir = tests_dir.resolve()

        failing, culprit = make_failing_suite(tests_dir, fail_collect)

        with pytest.raises(subprocess.CalledProcessError) as einfo:
            tr.measure_suite_coverage(tests_dir=tests_dir, source_dir=Path('src'), isolate_tests=False)


@pytest.mark.parametrize("fail_collect", [True, False])
@pytest.mark.parametrize("absolute", [True, False])
def test_measure_suite_coverage_isolated(absolute, fail_collect):
    with tempfile.TemporaryDirectory(dir=Path('.')) as tests_dir:
        tests_dir = Path(tests_dir)
        if absolute:
            tests_dir = tests_dir.resolve()

        failing, culprit = make_failing_suite(tests_dir, fail_collect)

        tr.measure_suite_coverage(tests_dir=tests_dir, source_dir=Path('src'), isolate_tests=True)

        # FIXME check coverage
